@@ -5,12 +5,15 @@ import warnings
 from typing import List, Dict, Optional, Any, IO, Tuple, Union, Iterator
 from functools import wraps
 import requests
-from .utils import request, build_portal_params, build_script_params, filename_from_url
-from .const import API_PATH, PORTAL_PREFIX, FMSErrorCode
+from .utils import (request, build_portal_params, build_script_params,
+                    filename_from_url, PlaceholderDict)
+from .const import (PORTAL_PREFIX, FMSErrorCode, API_VERSIONS, API_PATH_PREFIX,
+                    API_PATH)
 from .exceptions import BadJSON, FileMakerError, RecordError
 from .record import Record
 from .foundset import Foundset
 
+
 class Server(object):
     """The server class provides easy access to the FileMaker Data API
 
@@ -21,7 +24,8 @@ class Server(object):
                     user='db user name',
                     password='db password',
                     database='db name',
-                    layout='db layout'
+                    layout='db layout',
+                    api_version='v1'
                    )
         fms.login()
         fms.get_record(1)
@@ -39,7 +43,9 @@ class Server(object):
                  data_sources: Optional[List[Dict]] = None,
                  verify_ssl: Union[bool, str] = True,
                  type_conversion: bool = False,
-                 auto_relogin: bool = False) -> None:
+                 auto_relogin: bool = False,
+                 proxies: Optional[Dict] = None,
+                 api_version: Optional[str] = None) -> None:
         """Initialize the Server class.
 
         Parameters
@@ -55,8 +61,10 @@ class Server(object):
         database : str
             Name of database without extension, e.g. Contacts
         layout : str
-            Layout to work with. Can be changed between calls by setting the layout attribute again,
-            e.g.: fmrest_instance.layout = 'new_layout'.
+            Layout to work with. Can be changed between calls by setting the
+            layout attribute again, e.g.: fmrest_instance.layout = 'my_layout',
+            or by passing it per call via the request_layout parameter to one
+            of the Server's methods.
         data_sources : list, optional
             List of dicts in formatj
                 [{'database': 'db_file', 'username': 'admin', 'password': 'admin'}]
@@ -78,6 +86,13 @@ class Server(object):
             If True, tries to automatically get a new token (re-login) when a
             request comes back with a 952 (invalid token) error. Defaults to
             False.
+        proxies : dict, optional
+            Pass requests through a proxy, configure like so:
+            { 'https': 'http://127.0.0.1:8080' }
+        api_version : str, optional
+            Configure which version of the data API should be queried (e.g. v1)
+            It is recommended to set the version explicitly to prevent
+            potential future breaking changes.
         """
 
         self.url = url
@@ -88,6 +103,15 @@ class Server(object):
         self.data_sources = [] if data_sources is None else data_sources
         self.verify_ssl = verify_ssl
         self.auto_relogin = auto_relogin
+        self.proxies = proxies
+
+        if not api_version:
+            warnings.warn('No api_version given. Defaulting to v1.')
+            self.api_version = 'v1'
+        elif api_version not in API_VERSIONS:
+            raise ValueError(f'Invalid API version. Choose one of {API_VERSIONS}')
+        else:
+            self.api_version = api_version
 
         self.type_conversion = type_conversion
         if type_conversion and not importlib.util.find_spec("dateutil"):
@@ -115,6 +139,22 @@ class Server(object):
             bool(self._token), self.database, self.layout
         )
 
+    def _get_api_path(self, resource: str,
+                      layout: Optional[str] = None) -> str:
+        resource_path = resource.split('.')
+        if len(resource_path) > 1:
+            if resource_path[0] == 'meta':
+                path = API_PATH['meta'][resource_path[1]]
+            else:
+                raise ValueError('Invalid API path')
+        else:
+            path = API_PATH[resource_path[0]]
+
+        request_layout = layout if layout else self.layout
+        return (API_PATH_PREFIX.format(version=self.api_version) +
+                path.format_map(PlaceholderDict(database=self.database,
+                                                layout=request_layout)))
+
     def _with_auto_relogin(f):
         @wraps(f)
         def wrapper(self, *args, **kwargs):
@@ -142,7 +182,7 @@ class Server(object):
         Note that OAuth is currently not supported.
         """
 
-        path = API_PATH['auth'].format(database=self.database, token='')
+        path = self._get_api_path('auth').format(token='')
         data = {'fmDataSource': self.data_sources}
 
         response = self._call_filemaker('POST', path, data, auth=(self.user, self.password))
@@ -156,25 +196,31 @@ class Server(object):
         Note: this method is also called by __exit__
         """
 
-	# token is expected in endpoint for logout
-        path = API_PATH['auth'].format(database=self.database, token=self._token)
+        # token is expected in endpoint for logout
+        path = self._get_api_path('auth').format(token=self._token)
 
-	# remove token, so that the Authorization header is not sent for logout
-	# (_call_filemaker() will update the headers)
+        # remove token, so that the Authorization header is not sent for logout
+        # (_call_filemaker() will update the headers)
         self._token = ''
         self._call_filemaker('DELETE', path)
 
         return self.last_error == FMSErrorCode.SUCCESS.value
 
-    def create(self, record: Record) -> Optional[int]:
-        """Shortcut to create_record method. Takes record instance and calls create_record."""
+    def create(self, record: Record,
+               request_layout: Optional[str] = None) -> Optional[int]:
+        """Shortcut to create_record method. Takes record instance and calls
+        create_record.
+        """
         # TODO: support for handling foundset instances inside record instance
-        return self.create_record(record.to_dict(ignore_portals=True, ignore_internal_ids=True))
+        return self.create_record(
+            record.to_dict(ignore_portals=True, ignore_internal_ids=True),
+            request_layout=request_layout)
 
     @_with_auto_relogin
     def create_record(self, field_data: Dict[str, Any],
                       portals: Optional[Dict[str, Any]] = None,
-                      scripts: Optional[Dict[str, List]] = None) -> Optional[int]:
+                      scripts: Optional[Dict[str, List]] = None,
+                      request_layout: Optional[str] = None) -> Optional[int]:
         """Creates a new record with given field data and returns new internal record id.
 
         Parameters
@@ -192,11 +238,13 @@ class Server(object):
                 {'TO::field': 'hello', 'TO::field2': 'world'},
                 {'TO::field': 'another record'}
             ]
+        request_layout : str, optional
+            Sets the layout name for this request. This takes precedence over
+            the value stored in the Server instance's layout attribute and is
+            useful if you have a shared Server instance in a multi-threaded
+            program.
         """
-        path = API_PATH['record'].format(
-            database=self.database,
-            layout=self.layout,
-        )
+        path = self._get_api_path('record', request_layout)
 
         request_data: Dict = {'fieldData': field_data}
         if portals:
@@ -212,15 +260,21 @@ class Server(object):
 
         return int(record_id) if record_id else None
 
-    def edit(self, record: Record, validate_mod_id: bool = False) -> bool:
-        """Shortcut to edit_record method. Takes (modified) record instance and calls edit_record"""
+    def edit(self, record: Record, validate_mod_id: bool = False,
+             request_layout: Optional[str] = None) -> bool:
+        """Shortcut to edit_record method. Takes (modified) record instance
+        and calls edit_record.
+        """
         mod_id = record.modification_id if validate_mod_id else None
-        return self.edit_record(record.record_id, record.modifications(), mod_id)
+        return self.edit_record(
+            record.record_id, record.modifications(), mod_id,
+            request_layout=request_layout)
 
     @_with_auto_relogin
     def edit_record(self, record_id: int, field_data: Dict[str, Any],
                     mod_id: Optional[int] = None, portals: Optional[Dict[str, Any]] = None,
-                    scripts: Optional[Dict[str, List]] = None) -> bool:
+                    scripts: Optional[Dict[str, List]] = None,
+                    request_layout: Optional[str] = None) -> bool:
         """Edits the record with the given record_id and field_data. Return True on success.
 
         Parameters
@@ -247,12 +301,14 @@ class Server(object):
             Example: {'prerequest': ['my_script', 'my_param']}
             Allowed types: 'prerequest', 'presort', 'after'
             List should have length of 2 (both script name and parameter are required.)
+        request_layout : str, optional
+            Sets the layout name for this request. This takes precedence over
+            the value stored in the Server instance's layout attribute and is
+            useful if you have a shared Server instance in a multi-threaded
+            program.
         """
-        path = API_PATH['record_action'].format(
-            database=self.database,
-            layout=self.layout,
-            record_id=record_id
-        )
+        path = self._get_api_path(
+            'record_action', request_layout).format(record_id=record_id)
 
         request_data: Dict = {'fieldData': field_data}
         if mod_id:
@@ -270,17 +326,20 @@ class Server(object):
 
         return self.last_error == FMSErrorCode.SUCCESS.value
 
-    def delete(self, record: Record) -> bool:
+    def delete(self, record: Record,
+               request_layout: Optional[str] = None) -> bool:
         """Shortcut to delete_record method. Takes record instance and calls delete_record."""
         try:
             record_id = record.record_id
         except AttributeError:
             raise RecordError('Not a valid record instance. record_id is missing.') from None
 
-        return self.delete_record(record_id)
+        return self.delete_record(record_id, request_layout=request_layout)
 
     @_with_auto_relogin
-    def delete_record(self, record_id: int, scripts: Optional[Dict[str, List]] = None):
+    def delete_record(self, record_id: int,
+                      scripts: Optional[Dict[str, List]] = None,
+                      request_layout: Optional[str] = None):
         """Deletes a record for the given record_id. Returns True on success.
 
         Parameters
@@ -292,12 +351,14 @@ class Server(object):
             Example: {'prerequest': ['my_script', 'my_param']}
             Allowed types: 'prerequest', 'presort', 'after'
             List should have length of 2 (both script name and parameter are required.)
+        request_layout : str, optional
+            Sets the layout name for this request. This takes precedence over
+            the value stored in the Server instance's layout attribute and is
+            useful if you have a shared Server instance in a multi-threaded
+            program.
         """
-        path = API_PATH['record_action'].format(
-            database=self.database,
-            layout=self.layout,
-            record_id=record_id
-        )
+        path = self._get_api_path(
+            'record_action', request_layout).format(record_id=record_id)
 
         params = build_script_params(scripts) if scripts else None
 
@@ -308,37 +369,53 @@ class Server(object):
     @_with_auto_relogin
     def get_record(self, record_id: int, portals: Optional[List[Dict]] = None,
                    scripts: Optional[Dict[str, List]] = None,
-                   layout: Optional[str] = None) -> Record:
+                   layout: Optional[str] = None,
+                   request_layout: Optional[str] = None,
+                   response_layout: Optional[str] = None) -> Record:
         """Fetches record with given ID and returns Record instance
 
         Parameters
         -----------
         record_id : int
-            The FileMaker record id. Be aware that record ids CAN change (e.g. in cloned databases)
+            The FileMaker record id. Be aware that record ids CAN change
+            (e.g. in cloned databases)
         portals : list
-            A list of dicts in format [{'name':'objectName', 'offset':1, 'limit':50}]
+            A list of dicts in format
+            [{'name':'objectName', 'offset':1, 'limit':50}]
 
-            Use this if you want to limit the amout of data returned. Offset and limit are optional
-            with default values of 1 and 50, respectively.
+            Use this if you want to limit the amout of data returned. Offset
+            and limit are optional with default values of 1 and 50,
+            respectively.
             All portals will be returned when portals==None. Default None.
         scripts : dict, optional
             Specify which scripts should run when with which parameters
             Example: {'prerequest': ['my_script', 'my_param']}
             Allowed types: 'prerequest', 'presort', 'after'
-            List should have length of 2 (both script name and parameter are required.)
-        layout : str, optional
-            Passing a layout name allows you to set the response (!) layout.
-            This is helpful, for example, if you want to limit the number of fields/portals being
-            returned and have a dedicated response layout.
+            List should have length of 2 (both script name and parameter are
+            required.)
+        layout : str, optional, DEPRECATED -> use response_layout
+        request_layout : str, optional
+            Sets the layout name for this request. This takes precedence over
+            the value stored in the Server instance's layout attribute and is
+            useful if you have a shared Server instance in a multi-threaded
+            program.
+        response_layout : str, optional
+            Set the response layout. This is helpful, for example, if you
+            want to limit the number of fields/portals being returned and have
+            a dedicated response layout.
         """
-        path = API_PATH['record_action'].format(
-            database=self.database,
-            layout=self.layout,
-            record_id=record_id
-        )
+        if layout is not None:
+            warnings.warn('layout parameter is deprecated and will be removed '
+                          'in the future. Please use response_layout instead.')
+
+        path = self._get_api_path('record_action',
+                                  request_layout).format(record_id=record_id)
 
         params = build_portal_params(portals, True) if portals else {}
-        params['layout.response'] = layout
+
+        # set response layout; layout param is only handled for backward-
+        # compatibility
+        params['layout.response'] = layout if layout else response_layout
 
         # build script param object in FMSDAPI style
         script_params = build_script_params(scripts) if scripts else None
@@ -347,13 +424,16 @@ class Server(object):
 
         response = self._call_filemaker('GET', path, params=params)
 
-        # pass response to foundset generator function. As we are only requesting one record though,
-        # we only re-use the code and immediately consume the first (and only) record via next().
+        # pass response to foundset generator function. As we are only
+        # requesting one record though, we only re-use the code and immediately
+        # consume the first (and only) record via next().
         return next(self._process_foundset_response(response))
 
     @_with_auto_relogin
     def perform_script(self, name: str,
-                       param: Optional[str] = None) -> Tuple[Optional[int], Optional[str]]:
+                       param: Optional[str] = None,
+                       request_layout: Optional[str] = None
+                       ) -> Tuple[Optional[int], Optional[str]]:
         """Performs a script with the given name and parameter.
 
         Returns tuple containing script error and result.
@@ -364,12 +444,14 @@ class Server(object):
             The script name as defined in FileMaker Pro
         param: str
             Optional script parameter
+        request_layout : str, optional
+            Sets the layout name for this request. This takes precedence over
+            the value stored in the Server instance's layout attribute and is
+            useful if you have a shared Server instance in a multi-threaded
+            program.
         """
-        path = API_PATH['script'].format(
-            database=self.database,
-            layout=self.layout,
-            script_name=name
-        )
+        path = self._get_api_path(
+            'script', request_layout).format(script_name=name)
 
         response = self._call_filemaker('GET', path, params={'script.param': param})
 
@@ -380,7 +462,8 @@ class Server(object):
         return script_error, script_result
 
     @_with_auto_relogin
-    def upload_container(self, record_id: int, field_name: str, file_: IO) -> bool:
+    def upload_container(self, record_id: int, field_name: str, file_: IO,
+                         request_layout: Optional[str] = None) -> bool:
         """Uploads the given binary data for the given record id and returns True on success.
         Parameters
         -----------
@@ -390,12 +473,15 @@ class Server(object):
             Name of the container field on the current layout without TO name. E.g.: my_container
         file_ : fileobj
             File object as returned by open() in binary mode.
+        request_layout : str, optional
+            Sets the layout name for this request. This takes precedence over
+            the value stored in the Server instance's layout attribute and is
+            useful if you have a shared Server instance in a multi-threaded
+            program.
         """
-        path = API_PATH['record_action'].format(
-            database=self.database,
-            layout=self.layout,
-            record_id=record_id
-        ) + '/containers/' + field_name + '/1'
+        path = self._get_api_path(
+            'record_action', request_layout).format(record_id=record_id)
+        path += '/containers/' + field_name + '/1'
 
         # requests library handles content type for multipart/form-data incl. boundary
         self._set_content_type(False)
@@ -408,7 +494,9 @@ class Server(object):
                     sort: Optional[List[Dict[str, str]]] = None,
                     portals: Optional[List[Dict[str, Any]]] = None,
                     scripts: Optional[Dict[str, List]] = None,
-                    layout: Optional[str] = None) -> Foundset:
+                    layout: Optional[str] = None,
+                    request_layout: Optional[str] = None,
+                    response_layout: Optional[str] = None) -> Foundset:
         """Requests all records with given offset and limit and returns result as
         (sorted) Foundset instance.
 
@@ -430,20 +518,30 @@ class Server(object):
             Example: {'prerequest': ['my_script', 'my_param']}
             Allowed types: 'prerequest', 'presort', 'after'
             List should have length of 2 (both script name and parameter are required.)
-        layout : str, optional
-            Passing a layout name allows you to set the response (!) layout.
-            This is helpful, for example, if you want to limit the number of fields/portals being
-            returned and have a dedicated response layout.
+        layout : str, optional, DEPRECATED -> use response_layout
+        request_layout : str, optional
+            Sets the layout name for this request. This takes precedence over
+            the value stored in the Server instance's layout attribute and is
+            useful if you have a shared Server instance in a multi-threaded
+            program.
+        response_layout : str, optional
+            Set the response layout. This is helpful, for example, if you
+            want to limit the number of fields/portals being returned and have
+            a dedicated response layout.
         """
-        path = API_PATH['record'].format(
-            database=self.database,
-            layout=self.layout
-        )
+        if layout is not None:
+            warnings.warn('layout parameter is deprecated and will be removed '
+                          'in the future. Please use response_layout instead.')
+
+        path = self._get_api_path('record', request_layout)
 
         params = build_portal_params(portals, True) if portals else {}
         params['_offset'] = offset
         params['_limit'] = limit
-        params['layout.response'] = layout
+
+        # set response layout; layout param is only handled for backward-
+        # compatibility
+        params['layout.response'] = layout if layout else response_layout
 
         if sort:
             params['_sort'] = json.dumps(sort)
@@ -464,7 +562,9 @@ class Server(object):
              offset: int = 1, limit: int = 100,
              portals: Optional[List[Dict[str, Any]]] = None,
              scripts: Optional[Dict[str, List]] = None,
-             layout: Optional[str] = None) -> Foundset:
+             layout: Optional[str] = None,
+             request_layout: Optional[str] = None,
+             response_layout: Optional[str] = None) -> Foundset:
         """Finds all records matching query and returns result as a Foundset instance.
 
         Parameters
@@ -495,23 +595,30 @@ class Server(object):
             Example: {'prerequest': ['my_script', 'my_param']}
             Allowed types: 'prerequest', 'presort', 'after'
             List should have length of 2 (both script name and parameter are required.)
-        layout : str, optional
-            Passing a layout name allows you to set the response (!) layout.
-            Your find will still be performed based on the Server.layout attribute.
-            This is helpful, for example, if you want to limit the number of fields/portals being
-            returned and have a dedicated response layout.
+        layout : str, optional, DEPRECATED -> use response_layout
+        request_layout : str, optional
+            Sets the layout name for this request. This takes precedence over
+            the value stored in the Server instance's layout attribute and is
+            useful if you have a shared Server instance in a multi-threaded
+            program.
+        response_layout : str, optional
+            Set the response layout. This is helpful, for example, if you
+            want to limit the number of fields/portals being returned and have
+            a dedicated response layout.
         """
-        path = API_PATH['find'].format(
-            database=self.database,
-            layout=self.layout
-        )
+        if layout is not None:
+            warnings.warn('layout parameter is deprecated and will be removed '
+                          'in the future. Please use response_layout instead.')
+
+        path = self._get_api_path('find', request_layout)
 
         data = {
             'query': query,
             'sort': sort,
             'limit': str(limit),
             'offset': str(offset),
-            'layout.response': layout
+            # "layout" param is only handled for backwards-compatibility
+            'layout.response': layout if layout else response_layout
         }
 
         # build script param object in FMSDAPI style
@@ -560,8 +667,8 @@ class Server(object):
         response = request(method='get',
                            url=file_url,
                            verify=self.verify_ssl,
-                           stream=stream
-                          )
+                           stream=stream,
+                           proxies=self.proxies)
 
         return (name,
                 response.headers.get('Content-Type'),
@@ -583,7 +690,7 @@ class Server(object):
             Example:
                 { 'Table::myField': 'whatever' }
         """
-        path = API_PATH['global'].format(database=self.database)
+        path = self._get_api_path('global')
 
         data = {'globalFields': globals_}
 
@@ -621,6 +728,88 @@ class Server(object):
             }
         return result
 
+    def get_product_info(self) -> Dict:
+        """Fetches product info and returns Dict instance
+
+        Parameters
+        -----------
+        none
+        """
+        path = self._get_api_path('meta.product')
+
+        response = self._call_filemaker('GET', path)
+
+        return response.get('productInfo', None)
+
+    def get_databases(self) -> Dict:
+        """Fetches database names and returns Dict instance
+
+        Parameters
+        -----------
+        none
+        """
+        path = self._get_api_path('meta.databases')
+
+        # https://fmhelp.filemaker.com/docs/18/en/dataapi/#get-metadata_get-database-names
+        # If Filter Databases in Client Applications is disabled:
+        # no Authorization header is required.
+        #
+        # If Filter Databases in Client Applications is enabled:
+        # Authorization: a base64-encoded string representing the account name
+        # and password to use to log in to the hosted database.
+
+        # See discussion here: https://github.com/davidhamann/python-fmrest/pull/46#issuecomment-1079328531
+        # "This [will] eventually overwrite the Authorization header with the data provided in the auth 
+        # argument (a comment there might be good as it could lead to confusion since we're adding an 
+        # Authorization header, but requests adds one as well)."
+
+        response = self._call_filemaker('GET', path, auth=(self.user, self.password))
+
+        return response.get('databases', None)
+
+    @_with_auto_relogin
+    def get_layouts(self) -> Dict:
+        """Fetches database layout names and returns Dict instance
+
+        Parameters
+        -----------
+        none
+        """
+        path = self._get_api_path('meta.layouts')
+
+        response = self._call_filemaker('GET', path)
+
+        return response.get('layouts', None)
+
+    @_with_auto_relogin
+    def get_scripts(self) -> Dict:
+        """Fetches database script names and returns Dict instance
+
+        Parameters
+        -----------
+        none
+        """
+        path = self._get_api_path('meta.scripts')
+
+        response = self._call_filemaker('GET', path)
+
+        return response.get('scripts', None)
+
+    @_with_auto_relogin
+    def get_layout(self, layout: Optional[str] = None) -> Dict:
+        """Fetches layout metadata and returns Dict instance
+
+        Parameters
+        -----------
+        none
+        """
+        target_layout = layout if layout else self.layout
+        path = self._get_api_path('meta.layouts') + f'/{target_layout}'
+
+        response = self._call_filemaker('GET', path)
+
+        return response.get('fieldMetaData', None)
+
     def _call_filemaker(self, method: str, path: str,
                         data: Optional[Dict] = None,
                         params: Optional[Dict] = None,
@@ -648,7 +837,8 @@ class Server(object):
         request_data = json.dumps(data) if data else None
 
         # if we have a token, make sure it's included in the header
-	# if not, the Authorization header gets removed (necessary for example for logout)
+        # if not, the Authorization header gets removed (necessary for example
+        # for logout)
         self._update_token_header()
 
         response = request(method=method,
@@ -657,8 +847,8 @@ class Server(object):
                            data=request_data,
                            verify=self.verify_ssl,
                            params=params,
-                           **kwargs
-                          )
+                           proxies=self.proxies,
+                           **kwargs)
 
         try:
             response_data = response.json()
